/* XPiezo - a tool to configure the X3D-Piezo Gyro
 * Copyright (C) 2006 Simon Budig <simon@budig.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include "xpiezo.h"

static void   xpiezo_config_load      (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_config_save      (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_config_load_x3d  (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_config_save_x3d  (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_preset_import    (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_preset_export    (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_select_preset    (GtkRadioAction *action,
                                       GtkRadioAction *current,
                                       XPiezoContext  *context);
static void   xpiezo_preset_reorder   (GtkAction      *action,
                                       XPiezoContext  *context);
static void   xpiezo_quit             (GtkAction      *action,
                                       XPiezoContext  *context);

static void   xpiezo_preset_im_export (XPiezoContext *context,
                                       gint           preset,
                                       gboolean       write);


const GtkActionEntry xpiezo_actions[] =
{
  { "xpiezo-file-menu", NULL, "_File", NULL, NULL, NULL },
  { "xpiezo-preset-menu", NULL, "_Preset", NULL, NULL, NULL },
  { "xpiezo-config-open", GTK_STOCK_OPEN, "Open",
    "<Control>O", "Open file with a complete X-3D configuration",
    G_CALLBACK (xpiezo_config_load) },
  { "xpiezo-config-save", GTK_STOCK_SAVE_AS, "Save",
    "<Control>S", "Save file with a complete X-3D configuration",
    G_CALLBACK (xpiezo_config_save) },
  { "xpiezo-config-defaults", GTK_STOCK_UNDO, "Defaults",
    "<Control>U", "Revert to default X-3D configuration",
    G_CALLBACK (xpiezo_config_defaults) },

  { "xpiezo-load-from-x3d", GTK_STOCK_YES, "Load from X-3D",
    "<Control>D", "Download configurtion from the X-3D",
    G_CALLBACK (xpiezo_config_load_x3d)},
  { "xpiezo-save-to-x3d", GTK_STOCK_NO, "Save to X-3D",
    "<Control>C", "Upload configuration to the X-3D",
    G_CALLBACK (xpiezo_config_save_x3d)},

  { "xpiezo-quit", GTK_STOCK_QUIT, "Quit",
    "<Control>Q", "Quit XPiezo", G_CALLBACK (xpiezo_quit) },

  { "xpiezo-preset-import", NULL, "Import preset",
    "<Control>I", "Import file to the current preset",
    G_CALLBACK (xpiezo_preset_import) },
  { "xpiezo-preset-export", NULL, "Export preset",
    "<Control>E", "Export the current preset to a file",
    G_CALLBACK (xpiezo_preset_export) },
  { "xpiezo-preset-reorder", NULL, "Reorder presets",
    "<Control>R", "Reorder the presets in this configuration",
    G_CALLBACK (xpiezo_preset_reorder) },

  { "xpiezo-toolbar", NULL, "Toolbar", NULL, NULL, NULL },
};

const GtkRadioActionEntry xpiezo_radio_actions[] =
{
  { "xpiezo-preset-1", NULL, "Preset 1", NULL, NULL, 0 },
  { "xpiezo-preset-2", NULL, "Preset 2", NULL, NULL, 1 },
  { "xpiezo-preset-3", NULL, "Preset 3", NULL, NULL, 2 },
  { "xpiezo-preset-4", NULL, "Preset 4", NULL, NULL, 3 },
};


GtkActionGroup *
xpiezo_get_actions (XPiezoContext *context)
{
  GtkActionGroup *action_group;

  action_group = gtk_action_group_new ("XPiezo");
  gtk_action_group_add_actions (action_group, xpiezo_actions,
                                G_N_ELEMENTS (xpiezo_actions), context);
  gtk_action_group_add_radio_actions (action_group, xpiezo_radio_actions,
                                      G_N_ELEMENTS (xpiezo_radio_actions),
                                      0, G_CALLBACK (xpiezo_select_preset),
                                      context);

  return action_group;
}


static void
xpiezo_preset_import (GtkAction     *action,
                      XPiezoContext *context)
{
  xpiezo_preset_im_export (context, context->cur_preset, FALSE);
}


static void
xpiezo_preset_export (GtkAction     *action,
                      XPiezoContext *context)
{
  xpiezo_preset_im_export (context, context->cur_preset, TRUE);
}


static void
xpiezo_config_load (GtkAction     *action,
                    XPiezoContext *context)
{
  xpiezo_preset_im_export (context, -1, FALSE);
}


static void
xpiezo_config_save (GtkAction     *action,
                    XPiezoContext *context)
{
  xpiezo_preset_im_export (context, -1, TRUE);
}


static void
xpiezo_preset_im_export (XPiezoContext *context,
                         gint           preset_no,  /* -1 for whole config */
                         gboolean       write)
{
  static GtkWidget *save_dialog = NULL;
  static GtkWidget *load_dialog = NULL;
  static gchar     *current_folder = NULL;

  GtkWidget *dialog;
  GtkWidget *warning_dialog = NULL;
  XPiezoPreset *preset;
  GError *error = NULL;
  char *filename;
  gint response;

  if (write)
    {
      if (!save_dialog)
        {
          GtkFileFilter *filefilter;
          filefilter = gtk_file_filter_new ();
          gtk_file_filter_add_pattern (filefilter, "*.pst");

          save_dialog = gtk_file_chooser_dialog_new
                                         ("Export Preset",
                                          GTK_WINDOW (context->main_window),
                                          GTK_FILE_CHOOSER_ACTION_SAVE,
                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                          NULL);
          gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (save_dialog),
                                                          TRUE);
          gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (save_dialog),
                                       filefilter);
        }

      gtk_window_set_title (GTK_WINDOW (save_dialog),
                            preset_no < 0 ? "Save Config" : "Export Preset");

      dialog = save_dialog;
    }
  else
    {
      if (!load_dialog)
        {
          GtkFileFilter *filefilter;
          filefilter = gtk_file_filter_new ();
          gtk_file_filter_add_pattern (filefilter, "*.pst");

          load_dialog = gtk_file_chooser_dialog_new
                                         ("Export Preset",
                                          GTK_WINDOW (context->main_window),
                                          GTK_FILE_CHOOSER_ACTION_OPEN,
                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                                          NULL);
          gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (load_dialog),
                                                          TRUE);
          gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (load_dialog),
                                       filefilter);
        }

      gtk_window_set_title (GTK_WINDOW (load_dialog),
                            preset_no < 0 ? "Load Config" : "Import Preset");

      dialog = load_dialog;
    }

  if (current_folder)
    {
      g_printerr ("restoring current folder to %s\n", current_folder);
      gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
                                           current_folder);
    }
  else
    {
      g_printerr ("no current folder to restore to\n");
    }

  response = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (response)
  {
    case GTK_RESPONSE_ACCEPT:
      filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      if (write)
        {
          if (preset_no < 0)
            {
              xpiezo_config_save_to_file (context, filename, &error);
            }
          else
            {
              preset = &context->presets[preset_no];
              xpiezo_preset_save_to_file (preset, filename, &error);
            }
        }
      else
        {
          if (preset_no < 0)
            {
              xpiezo_config_load_from_file (context, filename, &error);
            }
          else
            {
              XPiezoPreset tmp_preset;

              if (xpiezo_preset_load_from_file (&tmp_preset, filename,
                                                &error))
                context->presets[preset_no] = tmp_preset;
            }
        }

      if (current_folder)
        g_free (current_folder);

      current_folder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (dialog));
      g_printerr ("setting current folder to %s\n", current_folder);
      break;

    default:
      break;
  }

   gtk_widget_hide (GTK_WIDGET (dialog));

   if (error)
     {
       warning_dialog = gtk_message_dialog_new_with_markup
                             (GTK_WINDOW (context->main_window),
                              GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                              GTK_MESSAGE_WARNING,
                              GTK_BUTTONS_OK,
                              "<b>An error occured:</b>\n\n%s",
                              error->message);

       response = gtk_dialog_run (GTK_DIALOG (warning_dialog));

       gtk_widget_destroy (GTK_WIDGET (warning_dialog));
     }

  xpiezo_refresh_gui (context);
}


static void
xpiezo_config_io_x3d (XPiezoContext *context,
                      gboolean       write)
{
  static GtkWidget *dialog = NULL;
  gint response;
  gint finished = FALSE;

  if (!dialog)
    {
      dialog = gtk_message_dialog_new_with_markup
                          (GTK_WINDOW (context->main_window),
                           GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                           GTK_MESSAGE_INFO,
                           GTK_BUTTONS_CANCEL,
                           "Loading Config from X3D");
    }

  gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG (dialog),
                                 write ? "Writing config to X3D" :
                                         "Loading config from X3D");

  context->progress_dialog = dialog;

  if (write)
    xpiezo_x3d_init_config_write (context);
  else
    xpiezo_x3d_init_config_load (context);

  while (!(finished || context->error))
    {
      response = gtk_dialog_run (GTK_DIALOG (dialog));

      switch (response)
        {
          case GTK_RESPONSE_DELETE_EVENT:
          case GTK_RESPONSE_CANCEL:
              finished = TRUE;
              break;
          case XPIEZO_RESPONSE_PROGRESS:
              /* g_printerr ("%f\n", context->progress); */
              break;
          case XPIEZO_RESPONSE_FINISHED:
              finished = TRUE;
              break;
          default:
              break;
        }
    }

  if (context->error)
    g_printerr ("Error: %s\n", context->error->message);

  g_clear_error (&context->error);

  xpiezo_close_piezo (context, &context->error);

  if (context->error)
    g_printerr ("Error: %s\n", context->error->message);

  g_clear_error (&context->error);

  context->progress_dialog = NULL;
  context->read_callback = NULL;
  gtk_widget_hide (GTK_WIDGET (dialog));
}


static void
xpiezo_config_load_x3d (GtkAction     *action,
                        XPiezoContext *context)
{
  xpiezo_config_io_x3d (context, FALSE);
}

static void
xpiezo_config_save_x3d (GtkAction     *action,
                        XPiezoContext *context)
{
  xpiezo_config_io_x3d (context, TRUE);
}


static void
xpiezo_select_preset (GtkRadioAction *action,
                      GtkRadioAction *current,
                      XPiezoContext  *context)
{
  context->cur_preset = gtk_radio_action_get_current_value (current);
  gtk_combo_box_set_active (GTK_COMBO_BOX (context->preset_button),
                            context->cur_preset);

  xpiezo_refresh_gui (context);
}


static void
xpiezo_reorder_callback (GtkToggleButton *togglebutton,
                         gpointer         user_data)
{
  GtkWidget **rbuttons = (GtkWidget **) user_data;
  gboolean have_toggle;
  gint i, j, t0, t1;

  if (rbuttons[15] == NULL)
    /* still initializing... */
    return;

  if (!gtk_toggle_button_get_active (togglebutton))
    return;

  for (t0 = 0; t0 < 16; t0++)
    if (GTK_WIDGET (togglebutton) == rbuttons[t0])
      break;

  g_return_if_fail (0 <= t0 && t0 < 16);

  t1 = -1;

  /* search for the already toggeled button in this column */
  for (i = (t0 % 4); i < 16; i += 4)
    {
      if (i != t0 && gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (rbuttons[i])))
        {
          t1 = i;
          break;
        }
    }

  if (t1 < 0)
    return;   /* no conflict */

  for (i = 0; i < 4; i++)
    {
      have_toggle = FALSE;
      for (j = 0; j < 4; j++)
        {
          if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (rbuttons[i+j*4])))
            {
              have_toggle = TRUE;
              break;
            }
        }
      if (!have_toggle)
        break;
    }

  if (i < 4)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rbuttons[i + (t1/4)*4]),
                                  TRUE);
}


static void
xpiezo_preset_reorder (GtkAction     *action,
                       XPiezoContext *context)
{
  static GtkWidget *dialog = NULL;
  static GtkWidget *rbuttons[16] = { NULL, };
  XPiezoPreset copy_presets[4];
  gint i, j, response;

  if (!dialog)
    {
      GtkWidget *table, *label;
      GSList *rgroup;

      dialog = gtk_dialog_new_with_buttons
                         ("Move Presets",
                          GTK_WINDOW (context->main_window),
                          GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                          GTK_STOCK_CANCEL,
                          GTK_RESPONSE_REJECT,
                          GTK_STOCK_OK,
                          GTK_RESPONSE_ACCEPT,
                          NULL);

      table = g_object_new (GTK_TYPE_TABLE,
                            "n-rows", 6,
                            "n-columns", 6,
                            "visible", TRUE,
                            NULL);

      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
                          table,
                          TRUE, FALSE, 10);

      label = g_object_new (GTK_TYPE_LABEL,
                            "label", "<b>Move presets</b>",
                            "use-markup", TRUE,
                            "visible", TRUE,
                            NULL);

      gtk_table_attach (GTK_TABLE (table), label,
                        0, 1, 0, 1,
                        GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 10, 0);

      label = g_object_new (GTK_TYPE_LABEL,
                            "label", "<i>from</i>",
                            "xalign", 1.0,
                            "use-markup", TRUE,
                            "justify", GTK_JUSTIFY_RIGHT,
                            "visible", TRUE,
                            NULL);
      gtk_table_attach (GTK_TABLE (table), label,
                        0, 1, 1, 6,
                        GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 10, 0);

      label = g_object_new (GTK_TYPE_LABEL,
                            "label", "<i>to</i>",
                            "use-markup", TRUE,
                            "justify", GTK_JUSTIFY_CENTER,
                            "visible", TRUE,
                            NULL);
      gtk_table_attach (GTK_TABLE (table), label,
                        1, 6, 0, 1,
                        GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 10, 0);
      for (i = 0; i < 4; i++)
        {
          gchar *text;
          text = g_strdup_printf ("<small>%d</small>", i + 1);
          label = g_object_new (GTK_TYPE_LABEL,
                                "label", text,
                                "use-markup", TRUE,
                                "justify", GTK_JUSTIFY_RIGHT,
                                "visible", TRUE,
                                NULL);
          gtk_table_attach (GTK_TABLE (table), label,
                            1, 2, i+2, i+3,
                            0, GTK_EXPAND | GTK_FILL, 0, 0);
          label = g_object_new (GTK_TYPE_LABEL,
                                "label", text,
                                "use-markup", TRUE,
                                "justify", GTK_JUSTIFY_CENTER,
                                "visible", TRUE,
                                NULL);
          gtk_table_attach (GTK_TABLE (table), label,
                            i+2, i+3, 1, 2,
                            0, GTK_EXPAND | GTK_FILL, 0, 0);
          g_free (text);

          rgroup = NULL;
          for (j = 0; j < 4; j++)
            {
              rbuttons[i*4+j] = gtk_radio_button_new (rgroup);
              gtk_widget_show (rbuttons[i*4+j]);
              g_signal_connect (rbuttons[i*4+j], "toggled",
                                G_CALLBACK (xpiezo_reorder_callback),
                                &rbuttons);

              gtk_table_attach (GTK_TABLE (table), rbuttons[i*4+j],
                                j+2, j+3, i+2, i+3,
                                0, 0, 0, 0);
              rgroup = gtk_radio_button_get_group (GTK_RADIO_BUTTON (rbuttons[i*4+j]));
            }

        }
    }

  for (i = 0; i < 4; i++)
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rbuttons[i*4+i]), TRUE);

  response = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (response)
    {
      case GTK_RESPONSE_ACCEPT:
        for (i = 0; i < 4; i++)
          {
            for (j = 0; j < 4; j++)
              {
                if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (rbuttons[i*4+j])))
                  copy_presets[j] = context->presets[i];
              }
          }

        for (i = 0; i < 4; i++)
          context->presets[i] = copy_presets[i];

        xpiezo_refresh_gui (context);

        break;
      default:
        break;
    }

  gtk_widget_hide (GTK_WIDGET (dialog));

}


void
xpiezo_config_defaults (GtkAction     *action,
                        XPiezoContext *context)
{
  XPiezoPreset *preset;
  gint i, j;
  guint s_offset;

  for (j = 0; j < XPIEZO_PARAM_LAST; j++)
    {
      /* sanity check - did we keep the order in sync? */
      if (j == param_defs[j].param)
        {
          s_offset = param_defs[j].s_offset;
          for (i = 0; i < 4; i++)
            {
              preset = &context->presets[i];
              G_STRUCT_MEMBER (gint, preset, s_offset) = param_defs[j].factory[i];
            }
        }
      else
        {
          g_printerr ("Programming error: param_defs and enum not in sync\n");
          exit (1);
        }
    }

  xpiezo_refresh_gui (context);
}


static void
xpiezo_quit (GtkAction     *action,
             XPiezoContext *context)
{
  static GtkWidget *dialog = NULL;
  gint response;

  if (!dialog)
    dialog = gtk_message_dialog_new_with_markup
                         (GTK_WINDOW (context->main_window),
                          GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                          GTK_MESSAGE_WARNING,
                          GTK_BUTTONS_YES_NO,
                          "There are unsaved changes. Really quit?");

  response = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (response)
   {
     case GTK_RESPONSE_YES:
       gtk_main_quit ();
       break;
     default:
       gtk_widget_hide (GTK_WIDGET (dialog));
       break;
   }
}
