/* XPiezo - a tool to configure the X3D-Piezo Gyro
 * Copyright (C) 2006 Simon Budig <simon@budig.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>

#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "xpiezo.h"


static void
xpiezo_pst_from_preset (XPiezoPreset  preset,
                        gchar        *pst_binary,
                        gsize         n_bytes)
{
  guchar *pst = (guchar *) pst_binary;
  gint chksum, i;

  g_return_if_fail (n_bytes == 46);

#define LE_BYTES_FROM_GUINT32(val, x)  (x)[0] =  ((val) & 0xff),         \
                                       (x)[1] = (((val) >>  8) & 0xff),  \
                                       (x)[2] = (((val) >> 16) & 0xff),  \
                                       (x)[3] = (((val) >> 24) & 0xff)

  LE_BYTES_FROM_GUINT32 (preset.nr_kp,          pst +  0);
  LE_BYTES_FROM_GUINT32 (preset.nr_kd,          pst +  4);
  LE_BYTES_FROM_GUINT32 (preset.yaw_kp,         pst +  8);
  LE_BYTES_FROM_GUINT32 (preset.yaw_kd,         pst + 12);
  LE_BYTES_FROM_GUINT32 (preset.nr_k_stick,     pst + 16);
  LE_BYTES_FROM_GUINT32 (preset.yaw_k_stick,    pst + 20);
  LE_BYTES_FROM_GUINT32 (preset.nr_k_expo,      pst + 24);
  LE_BYTES_FROM_GUINT32 (preset.angle_max,      pst + 28);
  LE_BYTES_FROM_GUINT32 (preset.nr_k_expo_edge, pst + 32);
  LE_BYTES_FROM_GUINT32 (preset.k_fast_yaw,     pst + 36);

#undef LE_BYTES_FROM_GUINT32

  pst[40]  = preset.orientation << 4;
  pst[40] |= preset.fast_yaw ? 1 : 0;
  pst[41] = preset.fast_yaw_margin;
  pst[42] = preset.temp_comp;
  pst[43] = preset.nr_k_stick_direct;

  chksum = 0;
  for (i = 0; i < 44; i++)
    chksum += pst[i];

  pst[44] = chksum & 0xff;
  pst[45] = (chksum >> 8) & 0xff;

#if 0
  for (i = 0; i < 46; i++)
    g_printerr (i % 16 != 15 ? "%02x " : "%02x\n", pst[i]);

  if (i % 16 != 0)
    g_printerr ("\n");
#endif

}

gboolean
xpiezo_data_read (GIOChannel   *channel,
                  GIOCondition  condition,
                  gpointer      data)
{
  XPiezoContext *context = (XPiezoContext *) data;
  guint bytes_read, expected_bytes;

  if (!context->read_buf)
    {
      context->read_buf = g_malloc0 (2048);
      context->read_buf_size = 2048;
      context->bytes_read = 0;
    }

  g_io_channel_read_chars (channel,
                           context->read_buf + context->bytes_read,
                           context->read_buf_size - context->bytes_read,
                           &bytes_read, &context->error);

  if (context->error)
    g_printerr ("Error: %s\n", context->error->message);

#if 0
  g_printerr ("read: (%d) (%p)\n", condition, channel);
  for (i = 0; i < bytes_read; i++)
    g_printerr (i % 16 != 15 ? "%02x " : "%02x\n",
                (unsigned char) context->read_buf[i+context->bytes_read]);

  if (i % 16 != 0)
    g_printerr ("\n");

  g_printerr ("\n");
#endif

  context->bytes_read += bytes_read;

#if 0
  g_printerr ("Buffer:\n");
  for (i = 0; i < context->bytes_read; i++)
    g_printerr (i % 16 != 15 ? "%02x " : "%02x\n",
                (unsigned char) context->read_buf[i]);

  if (i % 16 != 0)
    g_printerr ("\n");
#endif


  if (context->bytes_read >= 3)
    {
      expected_bytes = -1;
      if (context->read_buf[0] == 'P' &&
          context->read_buf[1] == 'I')
        {
          switch (context->read_buf[2])
            {
              case 'v':
                expected_bytes = 10;
                break;
              case 'E':
                expected_bytes = 52;
                break;
              case 'h':
                expected_bytes = 12;
                break;
              case 'd':
                expected_bytes = 17;
                break;
              case 'o':
              case 'f':
                expected_bytes = 5;
                break;
              default:
                g_printerr ("unknown reply %d (%c)\n",
                            context->read_buf[2], context->read_buf[2]);
                break;
            }
        }

      if (expected_bytes < 0)
        {
          g_printerr ("error handling goes here!\n");
        }
      else
        {
          if (context->bytes_read >= expected_bytes)
            {
              data = g_memdup (context->read_buf, expected_bytes);
              g_memmove (context->read_buf,
                         context->read_buf + expected_bytes,
                         context->bytes_read - expected_bytes);
              context->bytes_read -= expected_bytes;

              context->read_callback (context, data, expected_bytes);
            }
        }
    }

  return TRUE;
}


gboolean
xpiezo_data_write_helper (GIOChannel   *channel,
                          GIOCondition  condition,
                          gpointer      data)
{
  XPiezoContext *context = (XPiezoContext *) data;
  gsize bytes_written;
  GError *error = NULL;

  g_io_channel_write_chars (channel,
                            context->write_buf + context->written,
                            context->write_n_bytes - context->written,
                            &bytes_written, &error);
  if (error)
    g_printerr ("write-error: %s\n", error->message);

  context->written += bytes_written;

  if (context->written < context->write_n_bytes)
    {
      return TRUE;
    }
  else
    {
      g_free (context->write_buf);
      context->write_buf = NULL;
      context->write_n_bytes = 0;
      context->written = 0;
      context->write_watch = 0;
      return FALSE;
    }
}


gboolean
xpiezo_data_write (XPiezoContext *context,
                   gchar         *data,
                   gint           n_bytes)
{
  if (context->written < context->write_n_bytes)
    return FALSE;

  g_return_val_if_fail (context->write_watch == 0, FALSE);

  if (context->write_buf)
    g_free (context->write_buf);

  context->write_buf = g_memdup (data, n_bytes);
  context->write_n_bytes = n_bytes;
  context->written = 0;

  if (context->write_watch)
    g_source_remove (context->write_watch);

  context->write_watch = g_io_add_watch (context->serial, G_IO_OUT,
                                         xpiezo_data_write_helper, context);

  return TRUE;
}


GIOChannel *
xpiezo_open_piezo (XPiezoContext *context,
                   gchar         *serial_name,
                   GError       **error)
{
  GError *tmp_error = NULL;
  struct termios tio;

  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
  g_return_val_if_fail (context->serial_name != NULL ||
                        serial_name != NULL, NULL);

  if (serial_name)
    {
      if (context->serial_name)
        g_free (context->serial_name);

      context->serial_name = g_strdup (serial_name);
    }

  context->serial = g_io_channel_new_file (context->serial_name,
                                           "r+", &tmp_error);

  if (tmp_error != NULL)
    {
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  context->serial_fd = g_io_channel_unix_get_fd (context->serial);

  if (tcgetattr (context->serial_fd, &tio) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  cfmakeraw (&tio);
  if (cfsetspeed (&tio, B38400) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  if (tcsetattr (context->serial_fd, TCSANOW, &tio) != 0)
    {
      g_io_channel_unref (context->serial);
      tmp_error = g_error_new (G_IO_CHANNEL_ERROR,
                               g_io_channel_error_from_errno (errno),
                               "%s",
                               g_strerror (errno));
      g_propagate_error (error, tmp_error);
      return NULL;
    }

  g_io_channel_set_encoding (context->serial, NULL, NULL);
  g_io_channel_set_buffered (context->serial, FALSE);
  context->read_watch = g_io_add_watch (context->serial, G_IO_IN,
                                        xpiezo_data_read, context);

  return context->serial;
}


void
xpiezo_close_piezo (XPiezoContext  *context,
                    GError        **error)
{
  g_return_if_fail (error == NULL || *error == NULL);
  g_return_if_fail (context->serial != NULL);

  if (context->read_watch)
    g_source_remove (context->read_watch);
  if (context->write_watch)
    g_source_remove (context->write_watch);
  context->read_watch = 0;
  context->write_watch = 0;

  if (context->read_buf)
    {
      g_free (context->read_buf);
      context->read_buf = NULL;
      context->read_buf_size = 0;
      context->bytes_read = 0;
    }

  g_io_channel_unref (context->serial);
  context->serial = NULL;
  context->serial_fd = -1;
}

static void
xpiezo_x3d_config_load_step (XPiezoContext *context,
                             gchar         *x3d_response,
                             gsize          len)
{
  gchar *cmd;
  gint i = 0;

  g_return_if_fail (len >= 3);

  cmd = g_strndup ("UFOPI\x00", 6);

  if (x3d_response[2] == 'v')
    {
      context->progress = 0.0;
      xpiezo_data_write (context, cmd, 6);
    }
  else if (x3d_response[2] == 'E')
    {
      i = (guchar) x3d_response[3];
      g_return_if_fail (0 <= i && i <= 3);

      context->tmp_presets[i] = xpiezo_preset_from_pst (x3d_response + 4, 46);

      if (i < 3)
        {
          i++;
          context->progress = i / 5.0;

          cmd[5] = (guchar) i;
          xpiezo_data_write (context, cmd, 6);
        }
      else
        {
          context->progress = 4.0 / 5.0;
          xpiezo_data_write (context, "UFOPIH", 6);
        }
      gtk_dialog_response (GTK_DIALOG (context->progress_dialog),
                           XPIEZO_RESPONSE_PROGRESS);
    }
  else if (x3d_response[2] == 'h')
    {
      for (i = 0; i < 4; i++)
        {
          context->tmp_presets[i].easyparams =
                         (x3d_response[3] & (1 << (4+i)) ? TRUE : FALSE);
          context->tmp_presets[i].weight =  (x3d_response[4+i] >> 2) & 0x07;
          context->tmp_presets[i].agility = (x3d_response[4+i] >> 5) & 0x07;
        }
      context->progress = 1.0;
      for (i = 0; i < 4; i++)
        context->presets[i] = context->tmp_presets[i];

      xpiezo_refresh_gui (context);

      gtk_dialog_response (GTK_DIALOG (context->progress_dialog),
                           XPIEZO_RESPONSE_FINISHED);
    }
  else
    {
      g_printerr ("unexpected response from X3D\n");
      gtk_dialog_response (GTK_DIALOG (context->progress_dialog),
                           GTK_RESPONSE_CANCEL);
    }

  g_free (cmd);
}


gboolean
xpiezo_x3d_init_config_load (XPiezoContext *context)
{
  context->read_callback = xpiezo_x3d_config_load_step;

  if (!context->serial)
    xpiezo_open_piezo (context, "/dev/ttyUSB0", &context->error);

  if (!context->error)
    xpiezo_data_write (context, "UFOPIV", 6);

  return (context->error) ? FALSE : TRUE;
}


static void
xpiezo_x3d_config_write_step (XPiezoContext *context,
                              gchar         *x3d_response,
                              gsize          len)
{
  static gint cfg_written = 0;
  gchar *header = "UFOBUFOPIW";
  gchar *cmd;
  gint i = 0;

  g_return_if_fail (len >= 3);

  cmd = g_malloc0 (57);
  for (i = 0; i < 10; i++)
    cmd[i] = header[i];

  if (x3d_response[2] == 'v')
    {
      context->progress = 0.0;
      cmd[10] = '\x00';
      xpiezo_pst_from_preset (context->presets[0], cmd + 11, 46);
      xpiezo_data_write (context, cmd, 57);
      cfg_written = 0;
    }
  else if (x3d_response[2] == 'o')
    {
      if (cfg_written < 3)
        {
          cfg_written++;
          context->progress = (cfg_written + 1) / 5.0;
          cmd[10] = (guchar) cfg_written;
          xpiezo_pst_from_preset (context->presets[cfg_written], cmd + 11, 46);

          xpiezo_data_write (context, cmd, 57);
        }
      else if (cfg_written == 3)
        {
          cfg_written = 4;
          context->progress = (cfg_written + 1) / 5.0;
          cmd[9] = 'X';
          cmd[10] = cmd[11] = cmd[12] = cmd[13] = 0;
          cmd[14] = cmd[15] = cmd[16] = cmd[17] = 0;

          for (i = 0; i < 4; i++)
            {
              cmd[10] |= (context->presets[i].easyparams ? 1 << (4 + i) : 0);
              cmd[11+i] = context->presets[i].weight << 2 |
                          context->presets[i].agility << 5 | i;
            }
          i = (guchar) cmd[10] + (guchar) cmd[11] + (guchar) cmd[12] +
              (guchar) cmd[13] + (guchar) cmd[14];
          cmd[16] = i & 0xff;
          cmd[17] = (i >> 8) & 0xff;

          xpiezo_data_write (context, cmd, 18);
        }
      else
        {
          cfg_written = 5;
          context->progress = 1.0;
        }

      gtk_dialog_response (GTK_DIALOG (context->progress_dialog),
                           cfg_written < 5 ? XPIEZO_RESPONSE_PROGRESS :
                                             XPIEZO_RESPONSE_FINISHED);
    }
  else
    {
      g_printerr ("unexpected response from X3D\n");
      gtk_dialog_response (GTK_DIALOG (context->progress_dialog),
                           GTK_RESPONSE_CANCEL);
    }

  g_free (cmd);
}


gboolean
xpiezo_x3d_init_config_write (XPiezoContext *context)
{
  context->read_callback = xpiezo_x3d_config_write_step;

  if (!context->serial)
    xpiezo_open_piezo (context, "/dev/ttyUSB0", &context->error);

  if (!context->error)
    xpiezo_data_write (context, "UFOPIV", 6);

  return (context->error) ? FALSE : TRUE;
}


gboolean
xpiezo_sensors_request (gpointer data)
{
  XPiezoContext *context = (XPiezoContext *) data;

  xpiezo_data_write (context, "UFOPID", 6);

  return TRUE;
}


void
xpiezo_preset_update_from_easy (XPiezoContext *context)
{
  XPiezoPreset *preset = &context->presets[context->cur_preset];
  gint weight, agility;

  if (!preset->easyparams)
    return;

  weight = preset->weight;
  agility = preset->agility;

  preset->nr_kp  = 150 + 14 * weight;
  preset->nr_kd  = 400 + 14 * weight;
  preset->yaw_kp = 380 +  6 * weight;
  preset->yaw_kd = 500 +  7 * weight;

  preset->nr_k_stick        =    8 +    2 * agility;
  preset->nr_k_expo         =   60 +   13 * agility;
  preset->nr_k_expo_edge    = 1000 + 2000 * agility;
  preset->nr_k_stick_direct =   10 + agility / 2;

  preset->yaw_k_stick       =   20 +    2 * agility;
  preset->k_fast_yaw        =   60 +   11 * agility;

  preset->angle_max = (agility < 2) ? 1000000 : 2000000;

  preset->fast_yaw_margin = 15;
  preset->temp_comp = 50;

  xpiezo_refresh_gui (context);
}


XPiezoPreset
xpiezo_preset_from_pst (gchar *pst_binary, gint n_bytes)
{
  XPiezoPreset preset = { 0, };
  guchar *pst = (guchar *) pst_binary;
  gint chksum, i;

  g_return_val_if_fail (n_bytes == 46, preset);

  chksum = 0;
  for (i = 0; i < 44; i++)
    chksum += pst[i];

  if (chksum != pst[44] + (pst[45] << 8))
    g_printerr ("Wrong checksum %d vs. %d\n", chksum, pst[44] + (pst[45] << 8));

#define GUINT32_FROM_LE_BYTES(x)  ((x)[0])        + \
                                 (((x)[1]) <<  8) + \
                                 (((x)[2]) << 16) + \
                                 (((x)[3]) << 24);

  preset.nr_kp          = GUINT32_FROM_LE_BYTES (pst +  0);
  preset.nr_kd          = GUINT32_FROM_LE_BYTES (pst +  4);
  preset.yaw_kp         = GUINT32_FROM_LE_BYTES (pst +  8);
  preset.yaw_kd         = GUINT32_FROM_LE_BYTES (pst + 12);
  preset.nr_k_stick     = GUINT32_FROM_LE_BYTES (pst + 16);
  preset.yaw_k_stick    = GUINT32_FROM_LE_BYTES (pst + 20);
  preset.nr_k_expo      = GUINT32_FROM_LE_BYTES (pst + 24);
  preset.angle_max      = GUINT32_FROM_LE_BYTES (pst + 28);
  preset.nr_k_expo_edge = GUINT32_FROM_LE_BYTES (pst + 32);
  preset.k_fast_yaw     = GUINT32_FROM_LE_BYTES (pst + 36);

#undef GUINT32_FROM_LE_BYTES

  preset.orientation       = pst[40] >> 4;
  preset.fast_yaw          = pst[40] % 2 ? TRUE : FALSE;
  preset.fast_yaw_margin   = pst[41];
  preset.temp_comp         = pst[42];
  preset.nr_k_stick_direct = pst[43];

  return preset;
}


void
xpiezo_keyfile_add_preset (GKeyFile     *keyfile,
                           gchar        *section_name,
                           XPiezoPreset *preset)
{
  GEnumClass *enum_class;
  GEnumValue *enum_value;
  gint i, s_offset;

  enum_class = g_type_class_ref (XPIEZO_TYPE_PARAMETER_TYPE);

  for (i = 0; i < XPIEZO_PARAM_LAST; i++)
    {
      enum_value = g_enum_get_value (enum_class, i);
      s_offset = param_defs[i].s_offset;

      if (i == XPIEZO_EASYPARAMS ||
          i == XPIEZO_YAW_FAST_YAW)
        {
          g_key_file_set_boolean (keyfile, section_name,
                                  enum_value->value_nick,
                                  G_STRUCT_MEMBER (gboolean, preset, s_offset));
        }
      else
        {
          g_key_file_set_integer (keyfile, section_name,
                                  enum_value->value_nick,
                                  G_STRUCT_MEMBER (gint, preset, s_offset));
        }
    }

  enum_value = g_enum_get_value (enum_class, XPIEZO_NR_KP);
  g_key_file_set_comment (keyfile, section_name, enum_value->value_nick,
                          "\n nick/roll parameters\n", NULL);
  enum_value = g_enum_get_value (enum_class, XPIEZO_YAW_KP);
  g_key_file_set_comment (keyfile, section_name, enum_value->value_nick,
                          "\n yaw parameters\n", NULL);

  g_type_class_unref (enum_class);
}


gboolean
xpiezo_preset_save_to_file (XPiezoPreset  *preset,
                            gchar         *filename,
                            GError       **error)
{
  GKeyFile   *keyfile;
  gchar      *section = "x3d-preset";
  gsize       size;
  gchar      *data;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  keyfile = g_key_file_new ();
  g_key_file_load_from_data (keyfile, "\n# Parameter set for the X-3D\n", -1,
                             G_KEY_FILE_KEEP_COMMENTS |
                             G_KEY_FILE_KEEP_TRANSLATIONS,
                             NULL);

  xpiezo_keyfile_add_preset (keyfile, section, preset);

  data = g_key_file_to_data (keyfile, &size, error);
  g_key_file_free (keyfile);

  if (*error)
    return FALSE;

  g_file_set_contents (filename, data, size, error);
  g_free (data);

  if (*error)
    return FALSE;

  return TRUE;
}


gboolean
xpiezo_config_save_to_file (XPiezoContext *context,
                            gchar         *filename,
                            GError       **error)
{
  GKeyFile   *keyfile;
  gchar      *section = NULL, *comment = NULL;
  gsize       size;
  gchar      *data;
  gint        i;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  keyfile = g_key_file_new ();
  g_key_file_load_from_data (keyfile,
                             "#################################\n"
                             "# Configuration for the X-3D\n\n", -1,
                             G_KEY_FILE_KEEP_COMMENTS |
                             G_KEY_FILE_KEEP_TRANSLATIONS,
                             NULL);

  g_key_file_set_integer (keyfile, "x3d-config", "n_presets", 4);

  for (i = 0; i < 4; i++)
    {
      section = g_strdup_printf ("x3d-preset-%d", i+1);
      comment = g_strdup_printf ("################################\n"
                                 " Preset No. %d\n",
                                 i+1);

      xpiezo_keyfile_add_preset (keyfile, section, &context->presets[i]);

      g_key_file_set_comment (keyfile, section, NULL, comment, NULL);
      g_free (comment);
      g_free (section);
    }

  data = g_key_file_to_data (keyfile, &size, error);
  g_key_file_free (keyfile);

  if (*error)
    return FALSE;

  g_file_set_contents (filename, data, size, error);
  g_free (data);

  if (*error)
    return FALSE;

  return TRUE;
}


gboolean
xpiezo_keyfile_load_preset (GKeyFile      *keyfile,
                            gchar         *section_name,
                            XPiezoPreset  *preset,
                            GError       **error)
{
  GEnumClass *enum_class;
  GEnumValue *enum_value;
  gint i, s_offset;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  enum_class = g_type_class_ref (XPIEZO_TYPE_PARAMETER_TYPE);

  for (i = 0; i < XPIEZO_PARAM_LAST; i++)
    {
      enum_value = g_enum_get_value (enum_class, i);
      s_offset = param_defs[i].s_offset;

      if (i == XPIEZO_EASYPARAMS ||
          i == XPIEZO_YAW_FAST_YAW)
        {
          G_STRUCT_MEMBER (gboolean, preset, s_offset) =
                   g_key_file_get_boolean (keyfile, section_name,
                                           enum_value->value_nick,
                                           error);
        }
      else
        {
          G_STRUCT_MEMBER (gint, preset, s_offset) =
                   g_key_file_get_integer (keyfile, section_name,
                                           enum_value->value_nick,
                                           error);
        }

      if (*error)
        break;
    }

  g_type_class_unref (enum_class);

  return *error ? FALSE : TRUE;
}


gboolean
xpiezo_preset_load_from_file (XPiezoPreset   *preset,
                              gchar          *filename,
                              GError        **error)
{
  GKeyFile    *keyfile;
  gchar       *data;
  const gchar *end = NULL;
  gsize        size;
  gboolean     err;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  keyfile = g_key_file_new ();

  if (!g_file_get_contents (filename, &data, &size, error))
    return FALSE;

  if (!(g_utf8_validate (data, size, &end) ||
        end - data == size))
    {
      g_set_error (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_UNKNOWN_ENCODING,
                   "invalid UTF-8 encoding at byte %d.", end - data);
      return FALSE;
    }

  if (!g_key_file_load_from_data (keyfile, data, size, G_KEY_FILE_NONE, error))
    return FALSE;

  err = xpiezo_keyfile_load_preset (keyfile, "x3d-preset", preset, error);

  g_key_file_free (keyfile);
  return err;
}


gboolean
xpiezo_config_load_from_file (XPiezoContext *context,
                              gchar         *filename,
                              GError       **error)
{
  XPiezoPreset tmp_presets[4];
  GKeyFile *keyfile;
  gboolean err = FALSE;
  gchar *section, *data;
  const gchar *end;
  gsize size;
  gint i;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  keyfile = g_key_file_new ();

  if (!g_file_get_contents (filename, &data, &size, error))
    return FALSE;

  if (!(g_utf8_validate (data, size, &end) ||
        end - data == size))
    {
      g_set_error (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_UNKNOWN_ENCODING,
                   "invalid UTF-8 encoding at byte %d.", end - data);
      return FALSE;
    }

  if (!g_key_file_load_from_data (keyfile, data, size, G_KEY_FILE_NONE, error))
    return FALSE;

  for (i = 0; i < 4; i++)
    {
      section = g_strdup_printf ("x3d-preset-%d", i+1);

      err = !xpiezo_keyfile_load_preset (keyfile, section,
                                         &tmp_presets[i], error);

      g_free (section);

      if (err)
        break;
    }

  g_key_file_free (keyfile);

  if (err)
    return FALSE;

  for (i = 0; i < 4; i++)
    context->presets[i] = tmp_presets[i];

  return TRUE;
}

#if 0
void
xpiezo_dump_default_table (XPiezoContext *context)
{
  GEnumClass *enum_class;
  guint       len, s_offset;
  gchar      *pst = NULL;

  enum_class = g_type_class_ref (XPIEZO_TYPE_PARAMETER_TYPE);

  g_file_get_contents ("neuer yaw/1.pst", &pst, &len, NULL);
  context->presets[0] = xpiezo_preset_from_pst ((guchar *) pst, len);
  g_file_get_contents ("neuer yaw/2.pst", &pst, &len, NULL);
  context->presets[1] = xpiezo_preset_from_pst ((guchar *) pst, len);
  g_file_get_contents ("neuer yaw/3.pst", &pst, &len, NULL);
  context->presets[2] = xpiezo_preset_from_pst ((guchar *) pst, len);
  g_file_get_contents ("neuer yaw/4.pst", &pst, &len, NULL);
  context->presets[3] = xpiezo_preset_from_pst ((guchar *) pst, len);

  for (len = 0; len < XPIEZO_PARAM_LAST; len++)
    {
      GEnumValue *slot_value;

      slot_value = g_enum_get_value (enum_class, len);

      s_offset = param_defs[len].s_offset;
      g_printerr ("%-20s: %8d, %8d, %8d, %8d\n",
                  slot_value->value_nick,
                  G_STRUCT_MEMBER (gint, &context->presets[0], s_offset),
                  G_STRUCT_MEMBER (gint, &context->presets[1], s_offset),
                  G_STRUCT_MEMBER (gint, &context->presets[2], s_offset),
                  G_STRUCT_MEMBER (gint, &context->presets[3], s_offset));
    }
}
#endif


int
main (int argc, char *argv[])
{
  XPiezoContext *context;

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  context = g_new0 (XPiezoContext, 1);

  context->main_window = xpiezo_create_window (context);
  gtk_widget_show (context->main_window);

  xpiezo_config_defaults (NULL, context);
  context->cur_preset = 0;

  gtk_main ();
  return 0;
}
